from typing import Optional

import bpy
import numpy as np

from meshemy.blender.shortcut import select_object, select_one_or_all


def planar_decimate_mesh(degrees: float = 1.0, mesh_object_name: Optional[str] = None) -> None:
    ob = select_object(mesh_object_name)

    bpy.ops.object.mode_set(mode="OBJECT")

    bpy.ops.object.modifier_add(type="DECIMATE")
    ob.modifiers["Decimate"].decimate_type = "DISSOLVE"
    ob.modifiers["Decimate"].angle_limit = np.deg2rad(degrees)

    bpy.ops.object.modifier_apply(modifier="Decimate")


def merge_close(tolerance: float = 0.01, mesh_object_name: Optional[str] = None) -> None:
    _ob = select_one_or_all(mesh_object_name)

    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.mesh.remove_doubles(threshold=tolerance)



import bmesh
import bpy
import numpy as np
import open3d as o3d
from pydantic_numpy import NDArray, NDArrayFp64

from meshemy.blender.shortcut import load_mesh_into_object
from meshemy.blender.workflows import select_object
from meshemy.utility.io import o3d_from_vertices_faces


def load_mesh_from_o3d(open3d_triangular_mesh: o3d.geometry.TriangleMesh, name: str = "new_object"):
    # https://b3d.interplanety.org/en/how-to-create-mesh-through-the-blender-python-api/
    blender_mesh = bpy.data.meshes.new(f"{name}_mesh")
    blender_mesh.from_pydata(
        np.asarray(open3d_triangular_mesh.vertices),
        (),
        np.asarray(open3d_triangular_mesh.triangles),
    )
    blender_mesh.update()

    return load_mesh_into_object(name, blender_mesh)


def vertices_and_faces(mesh_object_name: str) -> tuple[NDArrayFp64, NDArray]:
    _ob = select_object(mesh_object_name)

    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.mesh.quads_convert_to_tris(quad_method="BEAUTY")
    bpy.ops.object.mode_set(mode="OBJECT")

    bm = bmesh.new()
    bm.from_mesh(select_object(mesh_object_name).data)

    return (
        np.array([v.co.copy() for v in bm.verts], dtype=np.float64),
        np.array([[v.index for v in f.verts] for f in bm.faces]),
    )


def load_mesh_to_o3d(mesh_object_name: str) -> o3d.geometry.TriangleMesh:
    return o3d_from_vertices_faces(*vertices_and_faces(mesh_object_name))



from typing import Optional, Sequence

import bpy
import numpy as np
import open3d as o3d
from pydantic import BaseModel, FilePath

from meshemy.blender.constant import SUFFIX_TO_READER
from meshemy.blender.shortcut import load_mesh_into_object
from meshemy.blender.utils import (
    load_mesh_from_o3d,
    vertices_and_faces,
)
from meshemy.blender.workflows import merge_close, planar_decimate_mesh, select_object
from meshemy.cookbook.open3d import Open3dCookbook
from meshemy.utility.io import o3d_from_vertices_faces
from meshemy.utility.seal import seal_mesh


class BlenderCookbook(BaseModel):
    mesh_name: str

    def planar_decimate(self, degree_tol: float) -> None:
        planar_decimate_mesh(degree_tol, mesh_object_name=self.mesh_name)

    def merge_close(self, distance_tol: float) -> None:
        merge_close(distance_tol, mesh_object_name=self.mesh_name)

    def to_o3d(self, attempt_seal_insurance: bool = True) -> Open3dCookbook:
        vertices, faces = vertices_and_faces(mesh_object_name=self.mesh_name)
        if attempt_seal_insurance:
            vertices, faces = seal_mesh(vertices, faces)
        return Open3dCookbook(mesh=o3d_from_vertices_faces(vertices, faces))

    @classmethod
    def from_data(
        cls,
        name: str,
        vertices: Sequence,
        edges: Optional[Sequence],
        faces: Optional[Sequence],
    ) -> "BlenderCookbook":
        assert edges is not None or faces is not None

        mesh = bpy.data.meshes.new(f"{name}_mesh")
        mesh.from_pydata(
            np.asarray(vertices),
            np.asarray(edges) if edges is not None else None,
            np.asarray(faces) if faces is not None else None,
        )
        mesh.update()

        load_mesh_into_object(name, mesh)

        return cls(mesh_name=name)

    @classmethod
    def from_o3d(cls, mesh: o3d.geometry.TriangleMesh, name: str) -> "BlenderCookbook":
        _ob = load_mesh_from_o3d(mesh, name)
        return cls(mesh_name=name)

    @classmethod
    def from_file(cls, path: FilePath, name: str, *args, **kwargs) -> "BlenderCookbook":
        SUFFIX_TO_READER[path.suffix](*args, **kwargs)
        # TODO: Object not bound to name

        return cls(mesh_name=name)



import bpy

SUFFIX_TO_READER = {
    ".gltf": bpy.ops.import_scene.gltf,
    ".glb": bpy.ops.import_scene.gltf,
    # ".stl": bpy.ops.wm.stl_import,
    ".ply": bpy.ops.import_scene.ply,
    # ".obj": bpy.ops.wm.obj_import,
}



from typing import Any, Optional

import bpy


def empty_scene():
    bpy.ops.scene.new(type="EMPTY")


def load_mesh_into_object(name: str, mesh):
    # make object from mesh
    new_object = bpy.data.objects.new(name, mesh)

    # make collection
    new_collection = bpy.data.collections.new("new_collection")
    bpy.context.scene.collection.children.link(new_collection)

    # add object to scene collection
    new_collection.objects.link(new_object)

    return new_object


def select_object(name: str):
    ob = bpy.context.scene.objects[name]
    bpy.ops.object.select_all(action="DESELECT")  # Deselect all objects
    bpy.context.view_layer.objects.active = ob  # Make the cube the active object
    ob.select_set(True)
    return ob


def select_one_or_all(name: Optional[str] = None) -> Any:
    if name:
        return select_object(name)
    bpy.ops.mesh.select_all(action="SELECT")



import pymeshfix
from pydantic_numpy import NDArray


def seal_mesh(vertices: NDArray, faces: NDArray) -> tuple[NDArray, NDArray]:
    return pymeshfix.clean_from_arrays(vertices, faces, joincomp=True, verbose=False)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['meshemy', 'meshemy.blender', 'meshemy.cookbook', 'meshemy.utility']

package_data = \
{'': ['*']}

install_requires = \
['bpy>=3.4.0,<4.0.0',
 'numpy>=1.24.1,<2.0.0',
 'open3d==0.16.0',
 'pydantic-numpy>=1.3.0,<2.0.0',
 'pydantic>=1.10.4,<2.0.0',
 'pymeshfix>=0.16.2,<0.17.0']

setup_kwargs = {
    'name': 'meshemy',
    'version': '0.2.0',
    'description': 'Developer friendly suite for manipulating mesh',
    'long_description': '# Meshemy: Python toolbelt for manipulating mesh\nConsolidation package for manipulating mesh. Comes with cookbook models from each package\n\n## Installation\n```shell\npip install meshemy\n```\nUse it in your poetry package\n```shell\npoetry add meshemy\n```\n',
    'author': 'caniko',
    'author_email': 'canhtart@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.11.0',
}


setup(**setup_kwargs)



import open3d as o3d
from pydantic_numpy import NDArray, NDArrayFp64


def o3d_from_vertices_faces(vertices: NDArrayFp64, faces: NDArray) -> o3d.geometry.TriangleMesh:
    return o3d.geometry.TriangleMesh(
        o3d.pybind.utility.Vector3dVector(vertices),
        o3d.pybind.utility.Vector3iVector(faces),
    )



from pathlib import Path

import numpy as np
import open3d as o3d
from pydantic import BaseModel
from pydantic_numpy import NDArray

from meshemy.blender.utils import load_mesh_from_o3d
from meshemy.cookbook.blender import BlenderCookbook


class Open3dCookbook(BaseModel):
    mesh: o3d.geometry.TriangleMesh

    class Config:
        arbitrary_types_allowed = True

    def smoothen(self, iterations: int, copy: bool = False) -> None:
        result = self.mesh.filter_smooth_taubin(number_of_iterations=iterations)
        if copy:
            self.__class__(mesh=result)
        self.mesh = result

    @property
    def vertices(self) -> NDArray:
        return np.asarray(self.mesh.vertices)

    @property
    def faces(self) -> NDArray:
        return np.asarray(self.mesh.triangles)

    @property
    def triangles(self) -> NDArray:
        return self.faces

    @property
    def to_blender(self, name: str) -> BlenderCookbook:
        _ob = load_mesh_from_o3d(self.mesh, name)
        return BlenderCookbook(mesh_name=name)

    def save(self, path: Path, mesh_format: str = ".glb") -> None:
        open3d_triangular_mesh = (
            self.mesh.remove_duplicated_triangles()
            .remove_duplicated_vertices()
            .remove_degenerate_triangles()
            .remove_unreferenced_vertices()
        )
        o3d.io.write_triangle_mesh(
            str(path.with_suffix(mesh_format)),
            open3d_triangular_mesh,
            write_vertex_colors=False,
            write_triangle_uvs=False,
        )
